#ifdef HAVE_CONFIG_H
# include "config.h"     
#endif
#include <iostream>
#include"dune/common/mpihelper.hh" // An initializer of MPI
#include"dune/common/exceptions.hh" // We use exceptions
#include<dune/common/fvector.hh>
#include<dune/common/static_assert.hh>
#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>
#include<dune/istl/solvers.hh>
#include<dune/istl/preconditioners.hh>
#include<dune/istl/io.hh>
#include<dune/pdelab/common/function.hh>
#include<dune/pdelab/finiteelementmap/pk3dfem.hh>
#include<dune/pdelab/finiteelementmap/p1fem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include<dune/pdelab/gridfunctionspace/genericdatahandle.hh>
#include<dune/pdelab/gridfunctionspace/interpolate.hh>
#include<dune/pdelab/constraints/constraints.hh>
#include<dune/pdelab/constraints/constraintsparameters.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
#include<dune/pdelab/backend/istlvectorbackend.hh>
#include<dune/pdelab/backend/istlmatrixbackend.hh>
#include<dune/pdelab/backend/istlsolverbackend.hh> //parallel
#include<dune/pdelab/stationary/linearproblem.hh>
#include<dune/pdelab/backend/novlpistlsolverbackend.hh>
#include<dune/grid/common/gridinfo.hh>


//Alugrid:
#include<dune/grid/alugrid.hh>
#include<dune/grid/io/file/dgfparser/dgfalu.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>

#include"example02_bctype.hh"
#include"Poisson_operator.hh"
#define P_order 2
#define intorder 5


int main(int argc, char** argv)
{
  try{
    //Maybe initialize Mpi
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is dune-poisson." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
        <<" processes!"<<std::endl;
    
    MPI_Comm comm = helper.getCommunicator();

    typedef Dune::ALUSimplexGrid<3,3> Grid;
    Dune::GridPtr<Grid> gridptr("tetgen_sqrt3_series.dgf",comm);               
    Grid &grid = *gridptr;
    grid.globalRefine(1);
    grid.loadBalance();
    //grid.globalRefine(1);
    
    typedef Dune::ALUSimplexGrid<3,3>::LeafGridView GV; 
    const GV& gv=grid.leafView();
    if (helper.rank() == 0)
    {
      Dune::gridinfo(grid);
    }     
      
 		if (helper.rank() == 0) std::cout << "Using Pk finite elements with k = " << P_order << std::endl;	  
      
    typedef GV::Grid::ctype Coord;
		typedef double Real;

		int DebugWait = 0;
		while (DebugWait);

		typedef Dune::PDELab::Pk3DLocalFiniteElementMap<GV,Coord,Real,P_order> FEM;
		FEM fem(gv);  
		
		typedef Dune::PDELab::NonoverlappingConformingDirichletConstraints CON;
		CON con;
		typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
		typedef Dune::PDELab::ISTLBCRSMatrixBackend<1,1> MBE;
		//typedef VBE::MatrixBackend::Matrix<double> TM;
		typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
		GFS gfs(gv,fem,con);			
		con.compute_ghosts(gfs);	// con stores indices of ghost dofs !!
		typedef Dune::PDELab::BackendVectorSelector<GFS,Real>::Type GFVec;	
		Dune::PDELab::ParallelISTLHelper<GFS> phelper(gfs);
		typedef Dune::PDELab::NonoverlappingScalarProduct<GFS, GFVec> NONSCAPA;
		NONSCAPA novlpSP(gfs,phelper);

		typedef BCType B;
		B bc;
		typedef GFS::ConstraintsContainer<Real>::Type CC;
		CC cc;
		cc.clear();
		Dune::PDELab::constraints(bc,gfs,cc,false); // Fill constraints container;
		std::cout << "Proc " << helper.rank() << " : Set up gfs of size " << gfs.globalSize() 
					<< " with constrained dofs = " << cc.size() << " ." << std::endl;


  	typedef PoissonLocalOperator<B, GFVec> PLOP; 
		PLOP plop(bc,intorder);

		typedef Dune::PDELab::GridOperator<GFS,GFS,PLOP,MBE,double,double,double,CC,CC,true> PGOS;
		PGOS pgos(gfs,cc,gfs,cc,plop);	

		typedef PGOS::MatrixContainer<GFVec::ElementType>::Type POM;
		POM poissonmatrix(pgos);	
		POM poissonmatrix_safty(pgos);
		poissonmatrix = 0.0;
		poissonmatrix_safty = 0.0;
		if (helper.rank() == 0) std::cout << "Building PoissonMatrix" << std::endl;	
		GFVec zero(gfs,0.0);
		zero = 0.0;
		poissonmatrix = 0.0;
		pgos.jacobian(zero,poissonmatrix);
		poissonmatrix_safty += poissonmatrix;
		if (helper.rank() == 0)	std::cout << "PoissonMatrix build " << std::endl;


		typedef typename GFVec::BaseT VecType;
		typedef typename Dune::PDELab::CommSelector<96,Dune::MPIHelper::isFake>::type Comm;
		typedef typename POM::BaseT MatrixType;
		MatrixType& poisson_mat_xtd = poissonmatrix_safty.base();

/*		one = 1.0;
		VecType oneBase = one.base();
		poisson_mat_xtd.mv(oneBase,oneBase);
*/     
		typedef typename Dune::PDELab::BlockProcessor<GFS>::AMGVectorTypeSelector<GFVec>::Type VectorType;
		typedef Dune::PDELab::VertexExchanger<PGOS,MatrixType> Exchanger; 
		Exchanger exchanger(gfs.gridview());
		exchanger.getextendedmatrix(poisson_mat_xtd,phelper);
		exchanger.sumEntries(poisson_mat_xtd);

		Comm oocc(gfs.gridview().comm(),Dune::SolverCategory::nonoverlapping);
		phelper.createIndexSetAndProjectForAMG(poisson_mat_xtd, oocc);
		typedef Dune::SeqSSOR<MatrixType,VectorType,VectorType,1> Smoother;
		Smoother smoother(poisson_mat_xtd, 1, 1);
		typedef Dune::NonoverlappingSchwarzScalarProduct<VectorType,Comm> PSP;
		PSP psp(oocc);      
		typedef Dune::NonoverlappingSchwarzOperator<MatrixType,VectorType,VectorType,Comm> Operator;
		Operator oop (poisson_mat_xtd,oocc);

		typedef Dune::NonoverlappingBlockPreconditioner<Comm, Smoother> ParSmoother;
		ParSmoother parsmoother(smoother, oocc);
		int verb = 0;
		if (gfs.gridview().comm().rank()==0) verb=1;
		Dune::CGSolver<VectorType> Poisson_SSOR_CG_solver(oop,psp,parsmoother,1e-10,5000,verb);
		Dune::InverseOperatorResult Poisson_stat;
		if (helper.rank() == 0)	std::cout << "Poisson solver defined" << std::endl;	

		// **********************************************
		// 				SETUP TEST OF RESIDUAL VALUE
		// x initial vector
		// A Poisson-Matrix
		// b = A * x
		// Test if A * x = b can give initial x
		//************************************************
		
		GFVec x(gfs,1.0);
	  Dune::PDELab::set_constrained_dofs(cc,0.0,x); 
		GFVec b(gfs,0.0);
		poissonmatrix.mv(x,b);
		novlpSP.make_consistent(b);
	  //Dune::PDELab::set_constrained_dofs(cc,0.0,b); //safty reasons - constrained dofs should already be zero

		GFVec b_safe(gfs,0.0);
		b_safe = b;
	
		x = 0.0;
	  Poisson_SSOR_CG_solver.apply(x,b,Poisson_stat); 

	 	Dune::PDELab::set_constrained_dofs(cc,0.0,x); // Needed, else not correct Poisson afterwards ?!!!
	  // x should already be consistent	  
	  if (helper.rank() == 0) std::cout << "Poisson Problem solved: " << std::endl;


		// ***************************************************
		// TEST IF RESIDUUM OF POISSON PROBLEM IS ZERO
		// ***************************************************
	
		// 1)  Test if A x = b 
	
		GFVec help(gfs,0.0);
		poissonmatrix.mv(x,help);
	  novlpSP.make_consistent(help);
		help -= b_safe;
		double just_double = novlpSP.dot(help,help);
		if (helper.rank() == 0)
	  {
	 		std::cout << "Residuum = " << just_double << std::endl;
  	}	
  	
  	// 2) Test whether x = x(gfs,1.0);
		help = 1.0;
	 	Dune::PDELab::set_constrained_dofs(cc,0.0,help); 
		help -= x;
		just_double = novlpSP.dot(help,help);
		if (helper.rank() == 0)
	  {
	 		std::cout << "Difference to initial x =  " << just_double << std::endl;
  	}	
	

  	// 3) See whether x has same norm for calculation with different # processes
  	
    just_double = novlpSP.dot(x,x);
		if (helper.rank() == 0)
	  {
  		std::cout << "Norm of x = " << just_double << std::endl;
  		// SP should have same result on one as on two/many processes
  	}	

       
    return 0;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
